# cython: language_level=3, profile=False
from contextlib import contextmanager

IF UNAME_SYSNAME != 'Windows':
    import cysignals
    from cysignals.signals cimport sig_off, sig_on

    @contextmanager
    def sig_on_off():
        sig_on()
        try:
            yield
        finally:
            sig_off()
ELSE:
    @contextmanager
    def sig_on_off():
        yield

from pyglui.cygl cimport glew as gl
from pyglui.cygl cimport utils
from pyglui.cygl.utils cimport RGBA
# pyfontstash needs to be import AFTER pyglui.cygl.glew
from pyglui.pyfontstash cimport fontstash as fs

__version__ = "1.31.0"
include 'gldraw.pxi'
include 'helpers.pxi'
include 'design_params.pxi'

import platform
from collections import namedtuple
from os import path
from time import time


cdef int UI_MOD_KEY  # defines modifier key for cut/copy/paste
if platform.system() == 'Darwin':
    UI_MOD_KEY = 8  # glfw.GLFW_MOD_SUPER
else:
    UI_MOD_KEY = 2  # glfw.GLFW_MOD_CONTROL

#global cdefs
cdef fs.Context glfont
cdef double ui_scale = 1.0
cdef bint should_redraw = True
cdef bint should_redraw_overlay = True

def get_roboto_font_path():
    return path.join(path.dirname(__file__),'Roboto-Regular.ttf')

def get_opensans_font_path():
    return path.join(path.dirname(__file__),'OpenSans-Regular.ttf')

def get_pupil_icons_font_path():
    return path.join(path.dirname(__file__),'pupil_icons.ttf')

def get_all_font_paths():
    return get_roboto_font_path(),get_opensans_font_path(),get_pupil_icons_font_path()

try:
    from six import unichr
except ImportError:
    pass
else:
    char = unichr


cdef class UI:
    '''
    The UI context for a glfw window.
    '''
    cdef Input new_input
    # cdef bint should_redraw
    cdef public list elements
    cdef FitBox window
    cdef fbo_tex_id ui_layer
    cdef fbo_tex_id overlay_layer

    def __cinit__(self):
        with sig_on_off():
            self.elements = []
            self.new_input = Input()
            self.window = FitBox(Vec2(0,0),Vec2(0,0))
            self.ui_layer = create_ui_texture(Vec2(200,200))
            self.overlay_layer = create_ui_texture(Vec2(200,200))

            #global init of gl fonts
            global glfont
            glfont = fs.Context()
            self.add_font('roboto',get_roboto_font_path() )
            self.add_font('opensans',get_opensans_font_path())
            self.add_font('pupil_icons',get_pupil_icons_font_path())

    def __init__(self):
        pass

    def add_font(self,font_name,font_path):
        if glfont:
            glfont.add_font(font_name,font_path)

    def terminate(self):
        global glfont
        glfont = None
        destroy_ui_texture(self.ui_layer)
        destroy_ui_texture(self.overlay_layer)
        self.elements = []

    def update_mouse(self,mx,my):
        if self.window.mouse_over(Vec2(mx,my)):
            self.new_input.dm.x,self.new_input.dm.y = mx-self.new_input.m.x, my-self.new_input.m.y
            self.new_input.m.x,self.new_input.m.y = mx,my


    def update_window(self,w,h):
        global should_redraw
        should_redraw = True
        self.window.size.x,self.window.size.y = w,h
        gl.glScissor(0,0,int(w),int(h))
        resize_ui_texture(self.ui_layer,self.window.size)
        resize_ui_texture(self.overlay_layer,self.window.size)

    def update_scroll(self, sx,sy):
        self.new_input.s.x = sx
        self.new_input.s.y = sy

    def update_key(self, key, scancode, action, mods):
        self.new_input.keys.append((key,scancode,action,mods))

    def update_char(self,c):
        try:
            self.new_input.chars.append(chr(c))
        except:
            pass

    def update_button(self,button,action,mods):
        self.new_input.buttons.append((button,action,mods))

    def update_clipboard(self, clipboard):
        self.new_input.cb = clipboard

    cdef sync(self):
        for e in self.elements:
            e.sync()

    cdef handle_input(self):
        if self.new_input:
            #let active elements deal with input first:
            for e in self.new_input.active_ui_elements:
                e.pre_handle_input(self.new_input)
            self.new_input.active_ui_elements = []

            #now everybody
            # ontop level in reverse so that menus drawn above other take precedence
            for e in self.elements[::-1]:
                e.handle_input(self.new_input,True)

            unused = Unused_Input(self.new_input.buttons, self.new_input.keys,
                                  self.new_input.chars, self.new_input.cb)
            self.new_input.purge()
        else:
            unused = Unused_Input([], [], [], self.new_input.cb)

        return unused

    cdef draw(self):
        global should_redraw
        global should_redraw_overlay
        global window_size
        window_size = self.window.size

        if should_redraw:
            should_redraw = False
            should_redraw_overlay = True
            push_view(self.window.size)
            render_to_ui_texture(self.ui_layer)
            glfont.clear_state()
            glfont.set_font('opensans')
            glfont.set_size(int(ui_scale * text_size))
            glfont.set_color_float(color_text_default)
            glfont.set_blur(.1)
            glfont.set_align(fs.FONS_ALIGN_TOP)

            for e in self.elements:
                e.draw(self.window, nested=False)

            render_to_screen()
            pop_view()

        draw_ui_texture(self.ui_layer)

        if should_redraw or should_redraw_overlay:
            should_redraw_overlay = False
            push_view(self.window.size)
            render_to_ui_texture(self.overlay_layer)
            glfont.clear_state()
            glfont.set_font('opensans')
            glfont.set_size(int(ui_scale * text_size))
            glfont.set_color_float(color_text_default)
            glfont.set_blur(.1)
            glfont.set_align(fs.FONS_ALIGN_TOP)

            for e in self.elements:
                e.draw_overlay(self.window, nested=False)

            render_to_screen()
            pop_view()

        draw_ui_texture(self.overlay_layer)


    def update(self):
        unused_Input = self.handle_input()
        self.sync()
        self.draw()
        return unused_Input

    def collect_menus(self):
        for e in self.elements:
            try:
                e.collect_in_window(self.window)
            except AttributeError:
                pass

    ###special methods to make UI behave like a list
    def append(self, obj):
        if not issubclass(obj.__class__,UI_element):
            raise Exception("Can only append UI elements, not: '%s'"%obj )
        self.elements.append(obj)
        global should_redraw
        should_redraw = True

    def extend(self, objs):
        for obj in objs:
            if not issubclass(obj.__class__,UI_element):
                raise Exception("Can only append UI elements, not: '%s'"%obj )
        self.elements.extend(objs)
        global should_redraw
        should_redraw = True

    def remove(self, obj):
        del self.elements[self.elements.index(obj)]
        global should_redraw
        should_redraw = True


    #identical to base_menu method
    cdef get_submenu_config(self):
        '''
        Growing menus are sometimes emebedded in Other menues. We load their configurations recursively.
        '''
        cdef dict submenus = {}
        for e in self.elements:
            if isinstance(e,(Growing_Menu,Scrolling_Menu,Stretching_Menu,Container)):
                submenus[e.label] = submenus.get(e.label,[]) + [e.configuration] #we could have two submenues with same label so we use a list for each submenu label cotaining the conf dicts for each menu
        return submenus

    #identical to base_menu method
    cdef set_submenu_config(self,dict submenus):
        '''
        Growing menus are sometimes emebedded in Other menues. We save their configurations recursively.
        '''
        if submenus:
            for e in self.elements:
                if isinstance(e,(Growing_Menu,Scrolling_Menu,Stretching_Menu,Container)):
                    e.configuration = submenus.get(e.label,[{}]).pop(0) #pop of the first menu conf dict in the list.

    property configuration:
        def __get__(self):
            cdef dict submenus = self.get_submenu_config()
            return {'submenus':submenus}

        def __set__(self,new_conf):
            self.set_submenu_config(new_conf.get('submenus',{}))



    def __len__ (self):
        return len(self.elements)

    def __getitem__ (self,x):
        return self.elements[x]

    def __setitem__ (self,x,obj):
        self.elements[x] = obj

    def __delitem__ (self,x):
        del self.elements[x]


    def __contains__ (self,obj):
        return obj in self.elements

    property scale:
        def __get__(self):
            return ui_scale

        def __set__(self, float val):
            global ui_scale
            ui_scale = val

include 'ui_elements.pxi'
include 'custom.pxi'
include 'timelines.pxi'
include 'menus.pxi'

#below are classes used by menu and ui_elements.
cdef class Synced_Value:
    '''
    an element that has a synced value
    attributes will be accecd through the attribute context unless you supply a getter.
    '''
    cdef object attribute_context
    cdef bint use_dict, trigger_overlay_only
    cdef str attribute_name
    cdef object _value
    cdef object getter
    cdef object setter
    cdef object on_change

    def __cinit__(self,str attribute_name, object attribute_context = None, getter=None, setter=None, on_change=None, trigger_overlay_only=False):
        assert attribute_context is not None or getter is not None
        self.attribute_context = attribute_context

        if isinstance(attribute_context,dict):
            self.use_dict = True
        else:
            self.use_dict = False

        self.attribute_name = attribute_name
        self.getter = getter
        self.setter = setter
        self.on_change = on_change
        self.trigger_overlay_only = trigger_overlay_only

    def __init__(self,str attribute_name, object attribute_context = None, getter=None, setter=None, on_change=None, trigger_overlay_only=False):
        if self.attribute_context is not None:
            if self.use_dict:
                try:
                    _ = self.attribute_context[self.attribute_name]
                except KeyError:
                    raise Exception("Dict: '%s' has no entry '%s'"%(self.attribute_context,self.attribute_name))
            else:
                try:
                    _ = getattr(self.attribute_context,self.attribute_name)
                except KeyError:
                    raise Exception("'%s' has no attribute '%s'"%(self.attribute_context,self.attribute_name))
        self.sync()



    cdef sync(self):
        global should_redraw
        global should_redraw_overlay
        if self.getter is not None:
            val = self.getter()
            if val != self._value:
                self._value = val
                if self.trigger_overlay_only:
                    should_redraw_overlay = True
                else:
                    should_redraw = True
                if self.on_change is not None:
                    self.on_change(self.value)

        elif self.use_dict:
            if self._value != self.attribute_context[self.attribute_name]:
                self._value = self.attribute_context[self.attribute_name]
                if self.trigger_overlay_only:
                    should_redraw_overlay = True
                else:
                    should_redraw = True
                if self.on_change is not None:
                    self.on_change(self.value)

        elif self._value != getattr(self.attribute_context,self.attribute_name):
            self._value = getattr(self.attribute_context,self.attribute_name)
            if self.trigger_overlay_only:
                should_redraw_overlay = True
            else:
                should_redraw = True
            if self.on_change is not None:
                self.on_change(self.value)

    property value:
        def __get__(self):
            return self._value
        def __set__(self,new_val):
            if self.setter is not None:
                self.setter(new_val)
            elif self.use_dict:
                self.attribute_context[self.attribute_name] = new_val
            else:
                setattr(self.attribute_context,self.attribute_name,new_val)


cdef class Input:
    '''
    Holds accumulated user input collect during a frame.
    '''

    cdef public list keys,chars,buttons,active_ui_elements
    cdef public basestring cb
    cdef Vec2 dm,m,s

    def __cinit__(self):
        self.keys = []
        self.buttons = []
        self.chars = []
        self.active_ui_elements = []
        self.m = Vec2(0,0)
        self.dm = Vec2(0,0)
        self.s = Vec2(0,0)
        self.cb = ''

    def __init__(self):
        pass

    def __nonzero__(self):
        return bool(self.keys or self.chars or self.buttons or self.dm or self.s)

    def purge(self):
        self.keys = []
        self.buttons = []
        self.chars = []
        self.dm.x = 0
        self.dm.y = 0
        self.s.x = 0
        self.s.y = 0
        self.cb = ''

    def __str__(self):
        return 'Current Input: \n   Mouse pos  : %s\n   Mouse delta: %s\n   Scroll: %s\n   Buttons: %s\n   Keys: %s\n   Chars: %s' %(self.m,self.dm,self.s,self.buttons,self.keys,self.chars)

Unused_Input = namedtuple('Unused_Input', ['buttons', 'keys', 'chars', 'clipboard'])


DEF drag_threshold = 10
cdef class Draggable:
    '''
    A rectangle that can be dragged.
    Does not move itself but the drag vector is added to 'value'
    Mouse postions units are scaled by 1/ui_scale
    '''
    cdef FitBox outline
    cdef Vec2 touch_point,drag_accumulator
    cdef bint selected,zero_crossing,dragged,catch_input
    cdef Vec2 value
    cdef int arrest_axis
    cdef object click_cb

    def __cinit__(self,Vec2 pos, Vec2 size, Vec2 value, arrest_axis = 0,zero_crossing=True,click_cb = None,catch_input=True):
        self.outline = FitBox(pos,size)
        self.value = value
        self.selected = False
        self.touch_point = Vec2(0,0)
        self.drag_accumulator = Vec2(0,0)

        self.arrest_axis = arrest_axis
        self.zero_crossing = zero_crossing
        self.click_cb = click_cb
        self.dragged = False
        self.catch_input = catch_input

    def __init__(self,Vec2 pos, Vec2 size, Vec2 value, arrest_axis = 0,zero_crossing=True,click_cb = None,catch_input=True):
        pass

    cdef draw(self, FitBox parent_size,bint nested=True):
        self.outline.compute(parent_size)
        self.outline.sketch()


    cpdef pre_handle_input(self,Input new_input):
        # we need to check for new clicks as touch pads can isse
        # a button down that is not follow by a button up.
        for b in new_input.buttons:
            if b[1] == 1:
                if not self.outline.mouse_over(new_input.m):
                    self.selected = False


    cdef handle_input(self,Input new_input, bint visible):
        global should_redraw

        if self.selected and new_input.dm:
            self.value -= self.drag_accumulator
            self.drag_accumulator = (new_input.m-self.touch_point)
            self.drag_accumulator *= 1/ui_scale

            if (abs(self.drag_accumulator.x) + abs(self.drag_accumulator.y))  > drag_threshold:
                self.dragged  = True

            if self.arrest_axis == 1:
                self.drag_accumulator.x = 0
            elif self.arrest_axis == 2:
                self.drag_accumulator.y = 0

            if not self.zero_crossing:
                if self.value.x > 0 and self.value.x + self.drag_accumulator.x <= 0:
                    self.drag_accumulator.x = .001 - self.value.x
                elif self.value.x < 0 and self.value.x + self.drag_accumulator.x >= 0:
                    self.drag_accumulator.x = -.001 - self.value.x
                elif self.value.x ==0:
                    self.drag_accumulator.x = 0
                if self.value.y > 0 and self.value.y + self.drag_accumulator.y <= 0:
                    self.drag_accumulator.y = .001 - self.value.y
                elif self.value.y < 0 and self.value.y + self.drag_accumulator.y >= 0:
                    self.drag_accumulator.y = -.001 - self.value.y
                elif self.value.y ==0:
                    self.drag_accumulator.y = 0

            self.value += self.drag_accumulator

            should_redraw = True

        for b in new_input.buttons[:]:#we need to make a copy for remove to work as desired
            if b[1] == 1 and visible:
                if self.outline.mouse_over(new_input.m):
                    self.selected = True
                    self.dragged  = False
                    self.touch_point.x = new_input.m.x
                    self.touch_point.y = new_input.m.y
                    self.drag_accumulator = Vec2(0,0)
                    if self.catch_input:
                        new_input.buttons.remove(b)
            if self.selected and b[1] == 0:
                self.selected = False
                if self.click_cb and not self.dragged:
                    self.click_cb()

        if self.selected:
            new_input.active_ui_elements.append(self)

    cdef sync(self):
        pass

cdef class FitBox:
    '''
    A box that will fit itself into a context.
    Specified by rules for x and y respectively:
        size positive -> size from self.org
        size 0 -> span into parent context and lock it like this. If you want it draggable use -.001 or .001
        size negative -> make the box to up to size pixels to the parent container.
        position negative -> align to the opposite side of context
        position 0  -> span into parent context and lock it like this. If you want it draggable use -.001 or .001

    This is quite expressive but does have a limitation:
        You cannot design a box that is outside of the parent context.

    Its made of 4 Vec2
        "design_org" "design_size" define the rules for placement and size

        "org" and "size" are the computed results of the box
            fitted and translated by its parent context

    Vec2 min_size is optional.


    Vec2 size and org will be computed when calling .compute(context,[nested])
        the globl UI scale factor will be applied to these vectors.

    '''
    cdef Vec2 design_org,org,design_size,size,min_size

    def __cinit__(self,Vec2 position,Vec2 size, Vec2 min_size = Vec2(0,0)):
        self.design_org = Vec2(position.x,position.y)
        self.design_size = Vec2(size.x,size.y)
        self.min_size = Vec2(min_size.x,min_size.y)
        # The values below are just temporay
        # and will be overwritten by compute.
        self.org = Vec2(position.x,position.y)
        self.org *= ui_scale
        self.size = Vec2(size.x,size.y)
        self.size *=ui_scale


    def __init__(self,Vec2 position,Vec2 size, Vec2 min_size = Vec2(0,0)):
        pass

    cdef collapse(self):

        #object is positioned from left(resp. top) and sized from object org
        if self.design_org.x >= 0 and  self.design_size.x  > 0:
            self.design_size.x = self.min_size.x
        #object is positioned from right (resp. bottom) and sized from context size
        elif self.design_org.x < 0 and self.design_size.x <= 0:
            self.design_org.x = self.design_size.x - self.min_size.x
            # self.design_size.x = self.min_size.x
        #object is positioned from left (top) and sized from context size:
        elif self.design_org.x >= 0 and self.design_size.x <= 0:
            pass
        #object is positioned from right and sized from object org
        elif self.design_org.x < 0 and self.design_size.x > 0:
            self.design_size.x = self.min_size.x
        else:
            pass

        #object is positioned from left(resp. top) and sized from object org
        if self.design_org.y >= 0 and self.design_size.y  > 0:
            self.design_size.y = self.min_size.y
        #object is positions from right (resp. bottom) and sized from context size
        elif self.design_org.y < 0 and self.design_size.y <= 0:
            self.design_org.y = self.design_size.y -self.min_size.y
            # self.design_size.y = self.min_size.y
        #object is positioned from left (top) and sized from context size:
        elif self.design_org.y >= 0 and self.design_size.y <= 0:
            pass
        #object is positioned from right and sized from object org
        elif self.design_org.y < 0 and self.design_size.y > 0:
            self.design_size.y = self.min_size.y
        else:
            pass


    cdef inflate(self,FitBox target):

        #object is positioned from left(resp. top) and sized from object org
        if self.design_org.x >= 0 and  self.design_size.x  > 0:
            self.design_size.x = target.design_size.x
        #object is positioned from right (resp. bottom) and sized from context size
        elif self.design_org.x < 0 and self.design_size.x <= 0:
            self.design_org.x = target.design_org.x
            self.design_size.x = target.design_size.x
        #object is positioned from left (top) and sized from context size:
        elif self.design_org.x >= 0 and self.design_size.x <= 0:
            pass
        #object is positioned from right and sized from object org
        elif self.design_org.x < 0 and self.design_size.x > 0:
            self.design_size.x = target.design_size.x
        else:
            pass

        #object is positioned from left(resp. top) and sized from object org
        if self.design_org.y >= 0 and  self.design_size.y  > 0:
            self.design_size.y = target.design_size.y
        #object is positioned from right (resp. bottom) and sized from context size
        elif self.design_org.y < 0 and self.design_size.y <= 0:
            self.design_org.y = target.design_org.y
            self.design_size.y = target.design_size.y
        #object is positioned from left (top) and sized from context size:
        elif self.design_org.y >= 0 and self.design_size.y <= 0:
            pass
        #object is positioned from right and sized from object org
        elif self.design_org.y < 0 and self.design_size.y > 0:
            self.design_size.y = target.design_size.y
        else:
            pass


    cdef compute(self,FitBox context):
        cdef float overflow

        # all x
        if self.design_org.x >=0:
            self.org.x = self.design_org.x * ui_scale
        else:
            self.org.x = context.size.x+self.design_org.x * ui_scale #design org is negative - double subtraction
        if self.design_size.x > 0:
            # size is direcly specified
            self.size.x = self.design_size.x * ui_scale
        else:
            self.size.x = context.size.x - self.org.x + self.design_size.x * ui_scale #design size is negative - double subtraction

        self.size.x = max(self.min_size.x * ui_scale, self.size.x)
        if self.design_org.x < 0 and self.design_size.x <= 0:
            overflow = self.org.x + self.size.x - context.size.x - self.design_size.x * ui_scale
            if overflow > 0:
                self.org.x -= overflow

        # finally translate into scene by parent org
        self.org.x +=context.org.x


        if self.design_org.y >=0:
            self.org.y = self.design_org.y * ui_scale
        else:
            self.org.y = context.size.y+self.design_org.y * ui_scale #design size is negative - double subtraction
        if self.design_size.y > 0:
            # size is direcly specified
            self.size.y = self.design_size.y * ui_scale
        else:
            self.size.y = context.size.y - self.org.y + self.design_size.y * ui_scale #design size is negative - double subtraction


        self.size.y = max(self.min_size.y * ui_scale,self.size.y)
        if self.design_org.y < 0 and self.design_size.y <= 0:
            overflow = self.org.y + self.size.y - context.size.y - self.design_size.y * ui_scale
            if overflow > 0:
                self.org.y -= overflow

        # finally translate into scene by parent org
        self.org.y +=context.org.y


    property rect:
        def __get__(self):
            return self.org.x,self.org.y,self.size.x,self.size.y

    property ellipse:
        def __get__(self):
            return self.org.x+self.size.x/2,self.org.y+self.size.y/2, self.size.x,self.size.y

    property center:
        def __get__(self):
            return self.org.x+self.size.x/2,self.org.y+self.size.y/2

    cpdef bint mouse_over(self,Vec2 m):
        return self.org.x <= m.x <= self.org.x+self.size.x and self.org.y <= m.y <=self.org.y+self.size.y

    cpdef bint mouse_over_margin(self, Vec2 m, Vec2 margin):
        return self.org.x - margin.x <= m.x <= self.org.x + self.size.x + margin.x and self.org.y - margin.y <= m.y <= self.org.y + self.size.y + margin.y

    def __str__(self):
        return "FitBox:\n   design org: %s size: %s\n   comptd org: %s size: %s"%(self.design_org,self.design_size,self.org,self.size)

    cdef same_design(self,FitBox other):
        return bool(self.design_org == other.design_org and self.design_size == other.design_size)

    cdef design_distance(self,FitBox other):
        cdef float d = 0
        d += abs(self.design_size.x-other.design_size.x)
        d += abs(self.design_size.y-other.design_size.y)
        d += abs(self.design_org.x-other.design_org.x)
        d += abs(self.design_org.y-other.design_org.y)
        return d

    cdef sketch(self, RGBA color=RGBA(*rect_color_default)):
        rect(self.org, self.size, color)

    cdef copy(self):
        return FitBox( Vec2(*self.design_org), Vec2(*self.design_size), Vec2(*self.min_size) )

    cdef computed_copy(self):
        cdef FitBox box = self.copy()
        box.org = Vec2(*self.org[:])
        box.size = Vec2(*self.size[:])
        return box

    cdef has_area(self):
        return 1 < self.size.x*self.size.y


cdef class Vec2:
    cdef public float x,y

    def __cinit__(self,float x, float y):
        self.x = x
        self.y = y

    def __init__(self,x,y):
        pass

    def __nonzero__(self):
        return bool(self.x or self.y)

    def __add__(self,Vec2 other):
        return Vec2(self.x+other.x,self.y+other.y)


    def __iadd__(self,Vec2 other):
        self.x += other.x
        self.y += other.y
        return self

    def __mul__(self,float factor):
        return Vec2(self.x * factor, self.y * factor)

    def __imul__(self,float factor):
        self.x *= factor
        self.y *= factor
        return self

    def __matmul__(self, Vec2 other):
        return self.x * other.x + self.y * other.y

    def __sub__(self,Vec2 other):
        return Vec2(self.x-other.x,self.y-other.y)

    def __isub__(self,Vec2 other):
        self.x -= other.x
        self.y -= other.y
        return self

    def __truediv__(self, float divident):
        return Vec2(self.x / divident, self.y / divident)

    def __itruediv__(self,float divident):
        self.x /= divident
        self.y /= divident
        return self

    def __len__(self):
        return 2

    def __str__(self):
        return 'x: %s y: %s'%(self.x,self.y)

    def __richcmp__(self,Vec2 other,int op):
        '''
        <   0
        ==  2
        >   4
        <=  1
        !=  3
        >=  5
        '''
        if op == 2:
            return bool(self.x == other.x and self.y == other.y)
        else:
            return NotImplemented

    def __getitem__(self,idx):
        if isinstance(idx,slice):
            return (self.x,self.y)[idx]
        else:
            if idx==0:
                return self.x
            elif idx==1:
                return self.y
            raise IndexError()

    def __setitem__(self,idx,obj):
        if isinstance(idx,slice):
            self.x,self.y = obj[0], obj[1] #we should be more specific about the kind of slice
        else:
            if idx ==0:
                self.x = obj
            elif idx == 1:
                self.y == obj
            else:
                raise IndexError()




cdef inline push_view(Vec2 size):
    gl.glMatrixMode(gl.GL_PROJECTION)
    gl.glPushMatrix()
    gl.glLoadIdentity()
    gl.glOrtho(0, size.x, size.y, 0, -1, 1)
    gl.glMatrixMode(gl.GL_MODELVIEW)
    gl.glPushMatrix()
    gl.glLoadIdentity()

cdef inline pop_view():
    gl.glMatrixMode(gl.GL_PROJECTION)
    gl.glPopMatrix()
    gl.glMatrixMode(gl.GL_MODELVIEW)
    gl.glPopMatrix()

cdef inline rect(Vec2 org, Vec2 size, RGBA color):
    gl.glColor4f(color.r, color.g, color.b, color.a)
    gl.glBegin(gl.GL_POLYGON)
    gl.glVertex3f(org.x,org.y,0.0)
    gl.glVertex3f(org.x,org.y+size.y,0.0)
    gl.glVertex3f(org.x+size.x,org.y+size.y,0.0)
    gl.glVertex3f(org.x+size.x,org.y,0.0)
    gl.glEnd()

cdef inline rect_outline(Vec2 org, Vec2 size, float line_width, RGBA color):
    gl.glColor4f(color.r, color.g, color.b, color.a)
    gl.glLineWidth(line_width)
    gl.glBegin(gl.GL_LINE_LOOP)
    gl.glVertex3f(org.x - line_width, org.y - line_width,0.0)
    gl.glVertex3f(org.x - line_width,org.y+size.y + line_width,0.0)
    gl.glVertex3f(org.x+size.x + line_width,org.y+size.y + line_width,0.0)
    gl.glVertex3f(org.x+size.x + line_width,org.y - line_width,0.0)
    gl.glEnd()

cdef inline rect_corners(Vec2 org, Vec2 end, RGBA color):
    gl.glColor4f(color.r, color.g, color.b, color.a)
    gl.glBegin(gl.GL_POLYGON)
    gl.glVertex3f(org.x,org.y,0.0)
    gl.glVertex3f(org.x,end.y,0.0)
    gl.glVertex3f(end.x,end.y,0.0)
    gl.glVertex3f(end.x,org.y,0.0)
    gl.glEnd()

cdef inline tripple_h(Vec2 org, Vec2 size, RGBA color=RGBA(1.,1.,1.,.5)):
    gl.glColor4f(color.r, color.g, color.b, color.a)
    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x + 3*ui_scale         ,org.y+5*ui_scale,0)
    gl.glVertex3f(org.x + size.x- 3*ui_scale ,org.y+ 5*ui_scale,0)

    gl.glVertex3f(org.x + 3   *ui_scale      ,org.y+size.y/2,0)
    gl.glVertex3f(org.x + size.x -3 *ui_scale,org.y+size.y/2,0)

    gl.glVertex3f(org.x + 3 *ui_scale        ,org.y+size.y-5*ui_scale,0)
    gl.glVertex3f(org.x + size.x -3*ui_scale ,org.y+size.y-5*ui_scale,0)

    gl.glEnd()

cdef inline tripple_d(Vec2 org, Vec2 size):
    cdef Vec2 o = org,s=size-Vec2(9*ui_scale,9*ui_scale)
    gl.glColor4f(1,1,1,.5)
    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)


    gl.glVertex3f(o.x,o.y+s.y,0)
    gl.glVertex3f(o.x + s.x,o.y,0)

    gl.glVertex3f(o.x + s.x*.0,o.y+s.y*.33,0)
    gl.glVertex3f(o.x + s.x*.33,o.y+s.y*0,0)

    gl.glVertex3f(o.x + s.x*.66,o.y+s.y*1.,0)
    gl.glVertex3f(o.x + s.x*1.,o.y+s.y*.66,0)

    gl.glEnd()

cdef inline triangle_h(Vec2 org, Vec2 size, RGBA color):
    gl.glColor4f(color.r,color.g,color.b,color.a)

    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x + 3 *ui_scale        ,org.y+5*ui_scale,0)
    gl.glVertex3f(org.x + size.x- 3*ui_scale ,org.y+ 5*ui_scale,0)
    gl.glEnd()

    #gl.glBegin(gl.GL_POLYGON)
    #gl.glVertex3f(org.x + 0         ,org.y+size.y/2,0)
    #gl.glVertex3f(org.x + size.x/2. ,org.y+size.y-5,0)
    #gl.glVertex3f(org.x + size.x -0 ,org.y+size.y/2,0)
    #gl.glEnd()

    gl.glBegin(gl.GL_LINE_LOOP)
    gl.glVertex3f(org.x + 3 *ui_scale        ,org.y+size.y/2,0)
    gl.glVertex3f(org.x + size.x/2. ,org.y+size.y-5*ui_scale,0)
    gl.glVertex3f(org.x + size.x -3*ui_scale ,org.y+size.y/2,0)
    gl.glEnd()

cdef inline tripple_v(Vec2 org, Vec2 size):
    gl.glColor4f(1,1,1,.5)
    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x + 5*ui_scale         ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + 5*ui_scale         ,org.y+size.y-3*ui_scale,0)

    gl.glVertex3f(org.x + size.x/2  ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + size.x/2 ,org.y+size.y-3*ui_scale,0)

    gl.glVertex3f(org.x + size.x -5*ui_scale ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + size.x -5*ui_scale ,org.y+size.y-3*ui_scale,0)

    gl.glEnd()

cdef inline triangle_right(Vec2 org, Vec2 size):
    gl.glColor4f(1,1,1,.5)

    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x + size.x -5*ui_scale ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + size.x -5*ui_scale ,org.y+size.y-3*ui_scale,0)
    gl.glEnd()


    gl.glBegin(gl.GL_LINE_LOOP)
    gl.glVertex3f(org.x + size.x/2  ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + 5*ui_scale          ,org.y+size.y/2,0)
    gl.glVertex3f(org.x + size.x/2  ,org.y+size.y-3*ui_scale,0)
    gl.glEnd()

cdef inline triangle_left(Vec2 org, Vec2 size):
    gl.glColor4f(1,1,1,.5)

    gl.glLineWidth(1.6*ui_scale)
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x + 5 *ui_scale        ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + 5 *ui_scale        ,org.y+size.y-3*ui_scale,0)
    gl.glEnd()


    gl.glBegin(gl.GL_LINE_LOOP)
    gl.glVertex3f(org.x + size.x/2  ,org.y+3*ui_scale,0)
    gl.glVertex3f(org.x + size.x -5*ui_scale ,org.y+size.y/2,0)
    gl.glVertex3f(org.x + size.x/2 ,org.y+size.y-3*ui_scale,0)
    gl.glEnd()

cdef inline line(Vec2 org, Vec2 end, RGBA color):
    gl.glColor4f(color.r, color.g, color.b, color.a)
    gl.glLineWidth(1.5*ui_scale) #thinner lines sometimes dont show on certain hardware.
    gl.glBegin(gl.GL_LINES)
    gl.glVertex3f(org.x, org.y,0)
    gl.glVertex3f(end.x, end.y,0)
    gl.glEnd()

cdef inline FitBox draw_seek_handle(FitBox handle, RGBA color):
    ''' Draw the seek handle and return FitBox that corresponds to the draggable area
    '''
    rect(handle.org, handle.size, color)
    cdef FitBox drag = handle.computed_copy()
    drag.org.x -= 10 * drag.size.x
    drag.size.x *= 21
    drag.size.y += 20 * ui_scale
    return drag

cdef inline FitBox draw_trim_handle(FitBox handle, float opening, RGBA color):
    ''' Draw a handle and return FitBox that corresponds draggable area
    '''
    # cdef Vec2 line_size = Vec2(2* ui_scale, handle.size.y * 3 / 4)
    # cdef Vec2 line_org = Vec2(0., handle.org.y + handle.size.y / 2 - line_size.y / 2)
    cdef tuple location
    if opening == 0.25:  # left handle
        location = handle.org.x + handle.size.x, handle.org.y + handle.size.y / 2
        # line_org.x = handle.org.x + handle.size.x - line_size.x
    elif opening == 0.75:  # right handle
        location = handle.org.x, handle.org.y + handle.size.y / 2
        # line_org.x = h andle.org.x
    else:
        location = handle.center
    utils.draw_progress(location, (opening + 0.25) % 1., (opening - 0.25) % 1.,
                        inner_radius=0., outer_radius=handle.size.y + ui_scale, color=color,
                        sharpness=0.9)
    cdef FitBox drag = handle.computed_copy()
    drag.org = Vec2(location[0] - drag.size.x, location[1] - drag.size.y)
    drag.size *= 2
    return drag

cdef inline FitBox draw_handle_top_right(Vec2 tip_loc, Vec2 handle_size, RGBA color):
    ''' Draw a handle and return FitBox that corresponds draggable area
    '''
    cdef float tip_length = 10.*ui_scale
    cdef float half_line_width = 2. * ui_scale

    rect_corners(Vec2(tip_loc.x - half_line_width, tip_loc.y),
                 Vec2(tip_loc.x + half_line_width, tip_loc.y - tip_length), color)

    rect_corners(Vec2(tip_loc.x - half_line_width + handle_size.x, tip_loc.y - tip_length),
                 Vec2(tip_loc.x - half_line_width, tip_loc.y - tip_length - handle_size.y), color)

    return FitBox(Vec2(tip_loc.x - half_line_width, tip_loc.y - tip_length - handle_size.y) / ui_scale, handle_size / ui_scale)

### OpenGL funtions for rendering to texture.
### Using this saves us considerable cpu/gpu time when the UI remains static.
### An almost identical implementation can be found in cylg.utils
ctypedef struct fbo_tex_id:
    gl.GLuint fbo_id
    gl.GLuint tex_id

cdef fbo_tex_id create_ui_texture(Vec2 tex_size):
    cdef fbo_tex_id ui_layer
    ui_layer.fbo_id = 0
    ui_layer.tex_id = 0

    # create Framebufer Object
    #requires gl ext or opengl > 3.0
    gl.glGenFramebuffers(1, &ui_layer.fbo_id)
    gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, ui_layer.fbo_id)

    #create texture object
    gl.glGenTextures(1, &ui_layer.tex_id)
    gl.glBindTexture(gl.GL_TEXTURE_2D, ui_layer.tex_id)
    # configure Texture
    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0,gl.GL_RGBA, int(tex_size.x),
                    int(tex_size.y), 0,gl.GL_RGBA, gl.GL_UNSIGNED_BYTE,
                    NULL)
    #set filtering
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)

    #attach texture to fbo
    gl.glFramebufferTexture2D(gl.GL_FRAMEBUFFER, gl.GL_COLOR_ATTACHMENT0,
                                gl.GL_TEXTURE_2D, ui_layer.tex_id, 0)

    if gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER) != gl.GL_FRAMEBUFFER_COMPLETE:
        raise Exception("UI Framebuffer could not be created.")

    #unbind fbo and texture
    gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
    gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)

    return ui_layer

cdef destroy_ui_texture(fbo_tex_id ui_layer):
    gl.glDeleteTextures(1,&ui_layer.tex_id)
    gl.glDeleteFramebuffers(1,&ui_layer.fbo_id)

cdef resize_ui_texture(fbo_tex_id ui_layer, Vec2 tex_size):
    gl.glBindTexture(gl.GL_TEXTURE_2D, ui_layer.tex_id)
    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0,gl.GL_RGBA, int(tex_size.x),
                    int(tex_size.y), 0,gl.GL_RGBA, gl.GL_UNSIGNED_BYTE,
                    NULL)
    gl.glBindTexture(gl.GL_TEXTURE_2D, 0)


cdef render_to_ui_texture(fbo_tex_id ui_layer):
    # set fbo as render target
    # blending method after:
    # http://stackoverflow.com/questions/24346585/opengl-render-to-texture-with-partial-transparancy-translucency-and-then-rende/24380226#24380226
    gl.glBlendFuncSeparate(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA,
                                gl.GL_ONE_MINUS_DST_ALPHA, gl.GL_ONE)
    gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, ui_layer.fbo_id)
    gl.glClearColor(0.,0.,0.,0.)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT)


cdef render_to_screen():
    # set rendertarget 0
    gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
    gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

cdef draw_ui_texture(fbo_tex_id ui_layer):
    # render texture

    # set blending
    gl.glBlendFunc(gl.GL_ONE, gl.GL_ONE_MINUS_SRC_ALPHA)

    # bind texture and use.
    gl.glBindTexture(gl.GL_TEXTURE_2D, ui_layer.tex_id)
    gl.glEnable(gl.GL_TEXTURE_2D)

    #set up coord system
    gl.glMatrixMode(gl.GL_PROJECTION)
    gl.glPushMatrix()
    gl.glLoadIdentity()
    gl.glOrtho(0, 1, 1, 0, -1, 1)
    gl.glMatrixMode(gl.GL_MODELVIEW)
    gl.glPushMatrix()
    gl.glLoadIdentity()

    gl.glColor4f(1.0,1.0,1.0,1.0)
    # Draw textured Quad.
    gl.glBegin(gl.GL_QUADS)
    gl.glTexCoord2f(0.0, 1.0)
    gl.glVertex2f(0,0)
    gl.glTexCoord2f(1.0, 1.0)
    gl.glVertex2f(1,0)
    gl.glTexCoord2f(1.0, 0.0)
    gl.glVertex2f(1,1)
    gl.glTexCoord2f(0.0, 0.0)
    gl.glVertex2f(0,1)
    gl.glEnd()

    #pop coord systems
    gl.glMatrixMode(gl.GL_PROJECTION)
    gl.glPopMatrix()
    gl.glMatrixMode(gl.GL_MODELVIEW)
    gl.glPopMatrix()

    gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
    gl.glDisable(gl.GL_TEXTURE_2D)
    gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)



from cpython.version cimport PY_MAJOR_VERSION

cimport pyglui.cygl.glew as gl


cdef str _to_str(object s):
    if PY_MAJOR_VERSION > 2:
        if type(s) is unicode:
            return s
        elif type(s) is bytearray:
            return s.decode('utf-8')
        else:
            return (<bytes>s).decode('utf-8')
    else:
        if type(s) is unicode:
            return s.encode('utf-8')
        else:
            return str(s)

cdef bytes _to_utf8_bytes(object s):
    if type(s) is unicode:
        return (<unicode>s).encode('utf-8')
    else:
        return <bytes>s


cdef class Shader:
    ''' Base shader class. '''

    def __cinit__(self, vertex_code = None, fragment_code = None, geometry_code = None):
        pass

    def __init__(self, vertex_code = None, fragment_code = None, geometry_code = None):
        '''
        Compile and link vertex code and fragment code into a shader.

        :Parameters:
            ``vertex_code``: string
                Vertex code

            ``fragment_code``: string
                Fragment code

            ``geometry_code``: string
                Geometry code
        '''

        self.uniforms = {}
        self._vertex_code = _to_utf8_bytes(vertex_code)
        self._fragment_code = _to_utf8_bytes(fragment_code)
        self._geometry_code = _to_utf8_bytes(geometry_code)

        # create the program handle
        self.handle = gl.glCreateProgram()

        # we are not linked yet
        self.linked = False

        # create the vertex shader
        self._build_shader(self._vertex_code, gl.GL_VERTEX_SHADER)

        # create the fragment shader
        self._build_shader(self._fragment_code, gl.GL_FRAGMENT_SHADER)

        # create the geometry shader
        self._build_shader(self._geometry_code, gl.GL_GEOMETRY_SHADER)

        # link the program
        self._link()


    cdef _get_shader_info(self,gl.GLuint shader):
        cdef gl.GLint log_length = 0
        gl.glGetShaderiv(shader,gl.GL_INFO_LOG_LENGTH,&log_length)
        cdef bytearray log = bytearray(log_length)
        gl.glGetShaderInfoLog(shader, log_length, NULL, log)
        return _to_str(log)


    cdef _get_program_info(self):
        cdef gl.GLint log_length = 0
        gl.glGetProgramiv(self.handle,gl.GL_INFO_LOG_LENGTH,&log_length)
        cdef bytearray log = bytearray(log_length)
        gl.glGetProgramInfoLog(self.handle, log_length, NULL, log)
        return log


    cdef _build_shader(self, const char * strings, shader_type):
        ''' Actual building of the shader '''

        count = len(strings)
        # if we have no source code, ignore this shader
        if count < 1:
            return

        # create the shader handle
        cdef gl.GLuint shader = gl.glCreateShader(shader_type)


        # Upload shader code we just have one string
        #void glShaderSource(GLhandle shaderObj, GLsizei count, const GLchar* const *string, const GLint *length)
        gl.glShaderSource(shader,1, &strings, NULL)

        # compile the shader
        gl.glCompileShader(shader)

        # retrieve the compile status
        cdef gl.GLint status = 0
        gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS, &status)

        # if compilation failed, raise exception and print the log
        if not status:
            if shader_type == gl.GL_VERTEX_SHADER:
                raise Exception('Vertex compilation: ' + self._get_shader_info(shader))
            elif shader_type == gl.GL_FRAGMENT_SHADER:
                raise Exception('Fragment compilation: ' + self._get_shader_info(shader))
            elif shader_type == gl.GL_GEOMETRY_SHADER:
                raise Exception('Geometry compilation: ' + self._get_shader_info(shader))
            else:
                raise Exception(self._get_shader_info(shader))
        else:
            # all is well, so attach the shader to the program
            gl.glAttachShader(self.handle, shader)

    cdef _link(self):
        ''' Link the program '''

        gl.glLinkProgram(self.handle)
        # retrieve the link status
        cdef gl.GLint status = 0
        gl.glGetProgramiv(self.handle, gl.GL_LINK_STATUS, &status)

        # if linking failed, print the log
        if not status:
            raise Exception('Linking: '+ self._get_program_info())
        else:
            # all is well, so we are linked
            self.linked = True

    cpdef bind(self):
        ''' Bind the program, i.e. use it. '''
        gl.glUseProgram(self.handle)

    cpdef unbind(self):
        ''' Unbind whatever program is currently bound - not necessarily this
            program, so this should probably be a class method instead. '''
        gl.glUseProgram(0)

    cpdef uniformf(self, str name, vals):
        ''' Uploads float uniform(s), program must be currently bound. '''

        loc = self.uniforms.get(name, gl.glGetUniformLocation(self.handle, _to_utf8_bytes(name)))
        #if loc < 0:
        #    raise ShaderException, \
        #        '''Unknow uniform location '%s' ''' % name
        self.uniforms[name] = _to_utf8_bytes(loc)

        cdef int val_len = len(vals)

        if val_len == 1:
            gl.glUniform1f(loc, vals[0])
        elif val_len ==2:
            gl.glUniform2f(loc, vals[0],vals[1])
        elif val_len ==3:
            gl.glUniform3f(loc, vals[0],vals[1],vals[2])
        elif val_len ==4:
            gl.glUniform4f(loc, vals[0],vals[1],vals[2],vals[3])

    cpdef uniform1f(self, str name, float val):
        ''' Upload float uniform, program must be currently bound. '''

        loc = self.uniforms.get(name, gl.glGetUniformLocation(self.handle, _to_utf8_bytes(name)))
        if loc < 0:
            raise Exception("Unknow uniform location '{}'".format(name))
        self.uniforms[name] = _to_utf8_bytes(loc)
        gl.glUniform1f(loc, val)

    cpdef uniform1i(self, str name, int val):
        ''' Upload integer uniform, program must be currently bound. '''

        loc = self.uniforms.get(name, gl.glGetUniformLocation(self.handle, _to_utf8_bytes(name)))
        if loc < 0:
            raise Exception("Unknow uniform location '{}'".format(name))
        self.uniforms[name] = loc
        gl.glUniform1i(loc, val)

    cpdef uniformi(self, str name, vals):
        ''' Upload integer uniform(s), program must be currently bound. '''

        loc = self.uniforms.get(name, gl.glGetUniformLocation(self.handle, _to_utf8_bytes(name)))
        #if loc < 0:
        #    raise ShaderException, \
        #        '''Unknow uniform location '%s' ''' % name
        self.uniforms[name] = loc

        cdef int val_len = len(vals)
        if val_len == 1:
            gl.glUniform1i(loc, vals[0])
        elif val_len ==2:
            gl.glUniform2i(loc, vals[0],vals[1])
        elif val_len ==3:
            gl.glUniform3i(loc, vals[0],vals[1],vals[2])
        elif val_len ==4:
            gl.glUniform4i(loc, vals[0],vals[1],vals[2],vals[3])


    cpdef uniform_matrixf(self, str name, float[:] mat):
        ''' Upload uniform matrix, program must be currently bound. '''

        loc = self.uniforms.get(name, gl.glGetUniformLocation(self.handle,_to_utf8_bytes(name)))
        #if loc < 0:
        #    raise ShaderException, \
        #        '''Unknow uniform location '%s' ''' % name
        self.uniforms[name] = loc

        # Upload the 4x4 floating point matrix
        gl.glUniformMatrix4fv(loc, 1, False, &mat[0])


    def get_vertex_code(self, lineno=True):
        code = ''
        for lineno, line in enumerate(_to_str(self._vertex_code).split('\n')):
            code += '{:3d}: {}\n'.format(lineno+1, line)
        return code

    def get_fragment_code(self,lineno=True):
        code = ''
        for lineno,line in enumerate(_to_str(self._fragment_code).split('\n')):
            code += '{:3d}: {}\n'.format(lineno+1, line)
        return code

    def get_geometry_code(self,lineno=True):
        code = ''
        for lineno,line in enumerate(_to_str(self._geometry_code).split('\n')):
            code += '{:3d}: {}\n'.format(lineno+1, line)
        return code



# cython: language_level=3

IF UNAME_SYSNAME == "Windows":
    cdef extern from 'Windows.h':
        pass
    cdef extern from '<GL/gl.h>':
        pass
ELSE:
    cdef extern from 'gl.h':
        pass

cdef extern from 'stb_truetype.h':
    pass

cdef extern from 'fontstash.h':

    enum: FONS_INVALID

    cdef enum FONSflags:
        FONS_ZERO_TOPLEFT
        FONS_ZERO_BOTTOMLEFT

    cdef enum FONSalign:
        FONS_ALIGN_LEFT
        FONS_ALIGN_CENTER
        FONS_ALIGN_RIGHT
        FONS_ALIGN_TOP
        FONS_ALIGN_MIDDLE
        FONS_ALIGN_BOTTOM
        FONS_ALIGN_BASELINE

    cdef enum FONSerrorCode:
        FONS_ATLAS_FULL
        FONS_SCRATCH_FULL
        FONS_STATES_OVERFLOW
        FONS_STATES_UNDERFLOW

    cdef struct FONSparams:
        int width
        int height
        unsigned char flags
        void *userPtr
        int (*renderCreate)(void *, int, int)
        int (*renderResize)(void *, int, int)
        void (*renderUpdate)(void *, int *, unsigned char *)
        void (*renderDraw)(void *, float *, float *, unsigned int *, int)
        void (*renderDelete)(void *)

    cdef struct FONSquad:
        float x0
        float y0
        float s0
        float t0
        float x1
        float y1
        float s1
        float t1

    cdef struct FONStextIter:
        float x
        float y
        float nextx
        float nexty
        float scale
        float spacing
        unsigned int codepoint
        short int isize
        short int iblur
        FONSfont *font
        int prevGlyphIndex
        char *str
        char *next
        char *end
        unsigned int utf8state

    cdef struct FONSfont:
        pass

    cdef struct FONScontext:
        pass

    FONScontext *fonsCreateInternal(FONSparams *params)

    void fonsDeleteInternal(FONScontext *s)

    void fonsSetErrorCallback(FONScontext *s, void (*callback)(void *, int, int), void *uptr)

    void fonsGetAtlasSize(FONScontext *s, int *width, int *height)

    int fonsExpandAtlas(FONScontext *s, int width, int height)

    int fonsResetAtlas(FONScontext *stash, int width, int height)

    int fonsAddFont(FONScontext *s, char *name, char *path)

    int fonsAddFontMem(FONScontext *s, char *name, unsigned char *data, int ndata, int freeData)

    int fonsGetFontByName(FONScontext *s, char *name)

    void fonsPushState(FONScontext *s)

    void fonsPopState(FONScontext *s)

    void fonsClearState(FONScontext *s)

    void fonsSetSize(FONScontext *s, float size)

    void fonsSetColor(FONScontext *s, unsigned int color)

    void fonsSetSpacing(FONScontext *s, float spacing)

    void fonsSetBlur(FONScontext *s, float blur)

    void fonsSetAlign(FONScontext *s, int align)

    void fonsSetFont(FONScontext *s, int font)

    float fonsDrawText(FONScontext *s, float x, float y, char *string, char *end)

    float fonsTextBounds(FONScontext *s, float x, float y, char *string, char *end, float *bounds)

    void fonsLineBounds(FONScontext *s, float y, float *miny, float *maxy)

    void fonsVertMetrics(FONScontext *s, float *ascender, float *descender, float *lineh)

    int fonsTextIterInit(FONScontext *stash, FONStextIter *iter, float x, float y, char *str, char *end)

    int fonsTextIterNext(FONScontext *stash, FONStextIter *iter, FONSquad *quad)

    unsigned char *fonsGetTextureData(FONScontext *stash, int *width, int *height)

    int fonsValidateTexture(FONScontext *s, int *dirty)

    void fonsDrawDebug(FONScontext *s, float x, float y)

cdef extern from 'glfontstash.h':

    FONScontext *glfonsCreate(int width, int height, int flags)
    void glfonsDelete(FONScontext* ctx)
    unsigned int glfonsRGBA(unsigned char r, unsigned char g, unsigned char b, unsigned char a)




cdef inline double lmap(double value, double istart, double istop, double ostart, double ostop):
    '''
    linear mapping of val from space 1 to space 2
    '''
    return ostart + (ostop - ostart) * ((value - istart) / (istop - istart))

cdef inline double clamp(double value, double minium, double maximum):
    return max(min(value,maximum),minium)

cdef inline double clampmap(double value, double istart, double istop, double ostart, double ostop):
    return clamp(lmap(value,istart,istop,ostart,ostop),ostart,ostop)

cdef inline bint mouse_over_center(Vec2 center, double w, double h, Vec2 m):
    return center.x-w/2 <= m.x <=center.x+w/2 and center.y-h/2 <= m.y <=center.y+h/2

cdef inline double step(double value, double start, double stop, double step):
    cdef double rest
    if step:
        value -=start
        rest = value%step
        #round down
        if rest < step/2.:
            value -=rest
        #round up
        else:
            value += step-rest
        value +=start
    return value

def frange(start, stop, step):
    '''
    todo: translate to cython function
    '''
    while start < stop:
        yield start
        start += step



"""
Script taken from: https://github.com/orlp/pygrafix
Appropriate Licence applies!
"""

import argparse
import os
import pathlib
import re


def generate_pxd(glew_header_loc, dest="."):
    with open(glew_header_loc) as fin:
        data = fin.read()

    # cython doesn't support const
    data = re.sub(r"\bconst\b", "", data)

    lines = data.split("\n")

    handled_lines = set()
    function_types = {}
    export_functions = {}
    function_defs = []
    enums = []

    # read in function types
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"typedef\s+([^(]+)\([^*]+\*\s*([a-zA-Z_][a-zA-Z0-9_]+)\)\s*(\(.+\))\s*;",
                line,
            )[0]
        except IndexError:
            continue

        function_types[result[1]] = (result[0].strip(), result[2])
        handled_lines.add(linenr)

    # read in exported functions
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"GLEW_FUN_EXPORT\s+([a-zA-Z_][a-zA-Z0-9_]+)\s+([a-zA-Z_][a-zA-Z0-9_]+)",
                line,
            )[0]
        except IndexError:
            continue

        export_functions[result[1]] = result[0]
        handled_lines.add(linenr)

    # match exported functions with function types
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"#define\s+([a-zA-Z_][a-zA-Z0-9_]+)\s+GLEW_GET_FUN\s*\(\s*([a-zA-Z_][a-zA-Z0-9_]+)\s*\)",
                line,
            )[0]
        except IndexError:
            continue

        export_func = export_functions[result[1]]
        function_defs.append(
            function_types[export_func][0]
            + " "
            + result[0]
            + function_types[export_func][1]
        )
        handled_lines.add(linenr)

    # add GLAPIENTRY functions
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"GLAPI\s+([a-zA-Z_][a-zA-Z0-9_]+)[^a-zA-Z_]+GLAPIENTRY[^a-zA-Z_]+([a-zA-Z_][a-zA-Z0-9_]+)\s*(\(.+\))\s*;",
                line,
            )[0]
        except IndexError:
            continue

        function_defs.append(" ".join(result))
        handled_lines.add(linenr)

    # read in numeric defines as enums
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"#define\s+([a-zA-Z_][a-zA-Z0-9_]+)\s+(?:(?:0x[0-9a-fA-F]+)|[0-9]+)",
                line,
            )[0]
        except IndexError:
            continue

        enums.append(result)
        handled_lines.add(linenr)

    # read in GLEW vars as enums
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"#define\s+([a-zA-Z_][a-zA-Z0-9_]+)\s+GLEW_GET_VAR\(.+\)", line
            )[0]
        except IndexError:
            continue

        enums.append(result)
        handled_lines.add(linenr)

    # also accept GL to GL defines as enums
    for linenr, line in enumerate(lines):
        try:
            result = re.findall(
                r"#define\s+(GL_[a-zA-Z0-9_]+)\s+GL_[a-zA-Z0-9_]+", line
            )[0]
        except IndexError:
            continue

        enums.append(result)
        handled_lines.add(linenr)

    pxdheader = """# cython: language_level=3
from libc.stdint cimport int64_t, uint64_t

cdef extern from "include_glew.h":
    ctypedef struct _cl_context:
        pass
    ctypedef struct _cl_event:
        pass
    ctypedef struct __GLsync:
        pass

    ctypedef unsigned short wchar_t
    ctypedef int ptrdiff_t

    ctypedef unsigned int GLenum
    ctypedef unsigned int GLbitfield
    ctypedef unsigned int GLuint
    ctypedef int GLint
    ctypedef int GLsizei
    ctypedef char GLchar
    ctypedef unsigned char GLboolean
    ctypedef signed char GLbyte
    ctypedef short GLshort
    ctypedef unsigned char GLubyte
    ctypedef unsigned short GLushort
    ctypedef unsigned long GLulong
    ctypedef float GLfloat
    ctypedef float GLclampf
    ctypedef double GLdouble
    ctypedef double GLclampd
    ctypedef int GLfixed
    ctypedef int GLclampx
    ctypedef void GLvoid

    ctypedef int64_t GLint64EXT
    ctypedef uint64_t GLuint64EXT
    ctypedef GLint64EXT GLint64
    ctypedef GLuint64EXT GLuint64
    ctypedef __GLsync *GLsync
    ctypedef char GLcharARB
    ctypedef ptrdiff_t GLintptr
    ctypedef ptrdiff_t GLsizeiptr
    ctypedef _cl_context *cl_context
    ctypedef _cl_event *cl_event
    ctypedef unsigned int GLhandleARB
    ctypedef ptrdiff_t GLintptrARB
    ctypedef ptrdiff_t GLsizeiptrARB
    ctypedef void* GLeglClientBufferEXT
    ctypedef unsigned short GLhalf
    ctypedef GLintptr GLvdpauSurfaceNV
    ctypedef long GLVULKANPROCNV

    ctypedef void *GLeglImageOES  # GL_EXT_EGL_image_storage

    ctypedef void (__stdcall *GLDEBUGPROCAMD)(GLuint id, GLenum category, GLenum severity, GLsizei length, GLchar *message, GLvoid *userParam)
    ctypedef void (__stdcall *GLDEBUGPROCARB)(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, GLchar *message, GLvoid *userParam)

    ctypedef void (__stdcall *GLDEBUGPROC)(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, GLvoid* userParam)
    ctypedef void (__stdcall *GLLOGPROCREGAL)(GLenum stream, GLsizei length, const GLchar *message, GLvoid *context)


    GLenum glewInit()
    GLboolean glewIsSupported(char *name)
    GLboolean glewIsExtensionSupported(char *name)
    GLboolean glewGetExtension(char* name)
    GLubyte *glewGetErrorString(GLenum error)
    GLubyte *glewGetString(GLenum name)


"""

    dest = pathlib.Path(dest)
    dest.mkdir(exist_ok=True, parents=True)
    with (dest / "glew.pxd").open("w") as fout:
        data = pxdheader

        data += "    enum:\n"
        data += "\n".join("        " + enum for enum in set(enums))
        data += "\n\n"

        def mod_func(func):
            keywords = [
                "and",
                "del",
                "for",
                "is",
                "raise",
                "assert",
                "elif",
                "from",
                "lambda",
                "return",
                "break",
                "else",
                "global",
                "not",
                "try",
                "class",
                "except",
                "if",
                "or",
                "while",
                "continue",
                "exec",
                "import",
                "pass",
                "yield",
                "def",
                "finally",
                "in",
                "print",
            ]

            # beautify functions
            func = re.sub(r"\s+", " ", func)  # collapse whitespace
            func = re.sub(r"\s*([()])\s*", r"\1", func)  # no whitespace near brackets
            func = re.sub(r"\s*,\s*", r", ", func)  # only whitespace __after__ comma
            func = re.sub(
                r"\s*(\*+)\s*", r" \1", func
            )  # beautify pointers in functions

            # cython doesn't support (void), need to do () for no arguments instead
            func = re.sub(r"\(void\)", "()", func)

            # keywords...
            for keyword in keywords:
                func = re.sub(r"\b%s\b" % keyword, keyword + "_", func)

            return func

        data += "\n".join("    " + mod_func(func) for func in function_defs)

        fout.write(data)

    with (dest / "unhandled_glew.h").open("w") as fout:
        data = "\n".join(
            lines[linenr] for linenr in range(len(lines)) if linenr not in handled_lines
        )
        data = re.sub("\n\n+", "\n", data)
        fout.write(data)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("glew_header_loc")
    parser.add_argument("destination")
    args = parser.parse_args()
    generate_pxd(args.glew_header_loc, dest=args.destination)



cimport cython
from libc.math cimport floor


cdef class Seek_Bar(UI_element):
    '''Seek bar that visualizes seek handles, trim marks and playback buttons

    Hover modes (descending priority):
        1   Seek handle is being hovered
        2   Right trim mark handle is being hovered
        3   Left trim mark handle is being hovered
        4   Seek bar is being hovered
        0   Seek bar is not being hovered
    '''

    cdef int recent_idx
    cdef double recent_ts, min_ts, mid_ts, max_ts
    cdef Vec2 point_click_seek_loc
    cdef readonly int hovering
    cdef FitBox bar, seek_handle, trim_left_handle, trim_right_handle
    cdef readonly bint seeking, trimming_left, trimming_right
    cdef Synced_Value trim_left, trim_right, playback_speed
    cdef object recent_idx_ts_getter, playback_time_setter, seeking_cb
    cdef Timeline_Menu handle_start_reference
    cdef Icon backwards, play, forwards

    def __cinit__(self, object sync_ctx,
                  double min_ts,
                  double max_ts,
                  object recent_idx_ts_getter,
                  object playback_time_setter,
                  object seeking_cb,
                  Timeline_Menu handle_start_reference,
                  *args, **kwargs):
        self.uid = id(self)
        self.trim_left = Synced_Value('trim_left_ts', sync_ctx, trigger_overlay_only=True)
        self.trim_right = Synced_Value('trim_right_ts', sync_ctx, trigger_overlay_only=True)
        self.playback_speed = Synced_Value('playback_speed', sync_ctx, trigger_overlay_only=True)
        self.recent_idx_ts_getter = recent_idx_ts_getter
        self.playback_time_setter = playback_time_setter
        self.seeking_cb = seeking_cb

        self.recent_ts = min_ts
        self.recent_idx = 0

        self.min_ts = min_ts
        self.max_ts = max_ts
        self.mid_ts = (min_ts + max_ts) / 2
        self.hovering = 0
        self.seeking = False
        self.trimming_left = False
        self.trimming_right = False
        self.handle_start_reference = handle_start_reference

        self.point_click_seek_loc = Vec2(0., 0.)
        self.outline = FitBox(Vec2(0., -50.), Vec2(0., 0.))
        self.bar = FitBox(Vec2(130., 21.), Vec2(-30., 3.))
        self.seek_handle = FitBox(Vec2(0., 0.), Vec2(0., 0.))
        self.trim_left_handle = FitBox(Vec2(0., 0.), Vec2(0., 0.))
        self.trim_right_handle = FitBox(Vec2(0., 0.), Vec2(0., 0.))

        play_icon = chr(0xE037)
        pause_icon = chr(0xe034)
        step_fwd_icon = chr(0xe044)
        step_bwd_icon = chr(0xe045)
        incr_pbs_icon = chr(0xE01F)  # pbs: playback speed
        decr_pbs_icon = chr(0xE020)

        def set_play(_):
            sync_ctx.play = not sync_ctx.play

        self.backwards = Icon('backwards', sync_ctx, label_font='pupil_icons',
                              label_getter=lambda: decr_pbs_icon if sync_ctx.play else step_bwd_icon,
                              getter=lambda: True,
                              hotkey=263)  # 263 = glfw.GLFW_KEY_LEFT
        self.forwards = Icon('forwards', sync_ctx, label_font='pupil_icons',
                             label_getter=lambda: incr_pbs_icon if sync_ctx.play else step_fwd_icon,
                             hotkey=262,  # 262 = glfw.GLFW_KEY_RIGHT
                             getter=lambda: True)

        self.play = Icon('play', sync_ctx, label_font='pupil_icons',
                         label_getter=lambda: pause_icon if sync_ctx.play else play_icon,
                         hotkey=32, # 32 = glfw.GLFW_KEY_SPACE
                         setter=set_play,
                         getter=lambda: True)

        self.backwards.outline = FitBox(Vec2(5, 0),Vec2(40, 40))
        self.play.outline = FitBox(Vec2(40, 0),Vec2(40, 40))
        self.forwards.outline = FitBox(Vec2(75, 0),Vec2(40, 40))

    def __init__(self, *args, **kwargs):
        pass

    cpdef sync(self):
        global should_redraw_overlay
        self.trim_left.sync()
        self.trim_right.sync()
        self.playback_speed.sync()
        self.backwards.sync()
        self.play.sync()
        self.forwards.sync()

        # test for value changes
        recent_idx, recent_ts = self.recent_idx_ts_getter()
        if self.recent_idx != recent_idx:
            should_redraw_overlay = True
        self.recent_idx, self.recent_ts = recent_idx, recent_ts

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        self.outline.compute(parent)
        self.outline.sketch(RGBA(0., 0., 0., 0.3))
        self.bar.compute(self.outline)
        self.bar.sketch(RGBA(1., 1., 1., 0.4))

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):

        self.backwards.draw(self.outline, nested=True, parent_read_only=False)
        self.play.draw(self.outline, nested=True, parent_read_only=False)
        self.forwards.draw(self.outline, nested=True, parent_read_only=False)

        cdef FitBox handle = FitBox(Vec2(0., 0.), Vec2(0., 0.))

        cdef double trim_left_val = self.trim_left.value
        cdef double trim_right_val = self.trim_right.value
        cdef double seek_x = clampmap(self.recent_ts, self.min_ts, self.max_ts, 0, self.bar.size.x)
        cdef double top_ext = self.handle_start_reference.element_space.org.y
        cdef double bot_ext = self.bar.org.y + self.bar.size.y + 20 * ui_scale
        cdef double selection_height = 5 * self.bar.size.y

        cdef double trim_l_x = clampmap(trim_left_val, self.min_ts, self.max_ts, 0, self.bar.size.x)
        handle.org.x = int(self.bar.org.x + trim_l_x - selection_height)
        handle.org.y = self.bar.org.y + self.bar.size.y / 2 - selection_height / 2
        handle.size.x = selection_height
        handle.size.y = selection_height
        self.trim_left_handle = draw_trim_handle(handle, 0.25, RGBA(*seekbar_trim_color_hover if self.hovering == 3 else seekbar_trim_color))

        cdef double trim_r_x = clampmap(trim_right_val, self.min_ts, self.max_ts, 0, self.bar.size.x)
        handle.org.x = int(self.bar.org.x + trim_r_x)
        handle.org.y = self.bar.org.y + self.bar.size.y / 2 - selection_height / 2
        self.trim_right_handle = draw_trim_handle(handle, 0.75, RGBA(*seekbar_trim_color_hover if self.hovering == 2 else seekbar_trim_color))

        # draw region between trim marks
        handle.size.x = handle.org.x - int(self.bar.org.x + trim_l_x)
        handle.org.x = int(self.bar.org.x + trim_l_x)
        handle.size.y = self.bar.size.y
        rect(handle.org, handle.size, RGBA(*seekbar_trim_color))

        handle.org.y += selection_height - handle.size.y
        rect(handle.org, handle.size, RGBA(*seekbar_trim_color))

        handle.org = Vec2(int(self.bar.org.x + seek_x - self.bar.size.y / 4), top_ext)
        handle.size = Vec2(self.bar.size.y / 2, bot_ext - top_ext)
        self.seek_handle = draw_seek_handle(handle, RGBA(*seekbar_seek_color_hover if self.hovering == 1 else seekbar_seek_color))

        if self.hovering == 4:
            utils.draw_points([self.point_click_seek_loc],
                              size=4*self.bar.size.y,
                              color=RGBA(*seekbar_seek_color_hover))

        # debug draggable areas
        # rect(self.seek_handle.org, self.seek_handle.size, RGBA(1., 0., 0., 0.2))
        # rect(self.trim_left_handle.org, self.trim_left_handle.size, RGBA(1., 0., 0., 0.2))
        # rect(self.trim_right_handle.org, self.trim_right_handle.size, RGBA(1., 0., 0., 0.2))

        cdef basestring speed_str = '{}x'.format(self.playback_speed.value)
        cdef basestring current_ts_str = self.format_ts(self.recent_ts)
        cdef basestring trim_left_str = self.format_ts(trim_left_val)
        cdef basestring trim_right_str = self.format_ts(trim_right_val)

        current_ts_str += ' – {:d}'.format(self.recent_idx)

        cdef double trim_num_offset = 3. * ui_scale
        cdef double time_y = self.bar.org.y - selection_height / 2 + 2*ui_scale
        cdef double nums_y = self.play.button.org.y + self.play.button.size.y - seekbar_number_size * ui_scale / 3
        # if self.hovering or self.seeking or self.trimming_left or self.trimming_right:
        glfont.push_state()
        glfont.set_font('opensans')
        glfont.set_size(seekbar_number_size * ui_scale)

        # draw actual text
        glfont.set_blur(.1)
        glfont.set_color_float((1., 1., 1., 1.))

        if self.recent_ts < self.mid_ts:
            glfont.set_align(fs.FONS_ALIGN_BOTTOM | fs.FONS_ALIGN_LEFT)
            glfont.draw_text(self.bar.org.x + seek_x + 5*ui_scale, time_y, current_ts_str)
        else:
            glfont.set_align(fs.FONS_ALIGN_BOTTOM | fs.FONS_ALIGN_RIGHT)
            glfont.draw_text(self.bar.org.x + seek_x - 5*ui_scale, time_y, current_ts_str)

        glfont.set_align(fs.FONS_ALIGN_TOP | fs.FONS_ALIGN_CENTER)
        # glfont.draw_text(self.seek_handle.org.x+self.seek_handle.size.x/2,
        #                  self.seek_handle.org.y+self.seek_handle.size.y + 3. * ui_scale,
        #                  speed_str)
        glfont.draw_text(self.play.button.center[0], nums_y, speed_str)

        glfont.set_align(fs.FONS_ALIGN_TOP | fs.FONS_ALIGN_RIGHT)
        glfont.draw_text(self.trim_left_handle.center[0] - trim_num_offset, nums_y, trim_left_str)

        glfont.set_align(fs.FONS_ALIGN_TOP | fs.FONS_ALIGN_LEFT)
        glfont.draw_text(self.trim_right_handle.center[0] + trim_num_offset, nums_y, trim_right_str)

        glfont.pop_state()

    @cython.cdivision(True)
    cdef format_ts(self, double ts):
        cdef double minutes, seconds
        ts -= self.min_ts
        minutes = floor(ts / 60)
        seconds = ts - (minutes * 60.)
        return '{:02.0f}:{:06.3f}'.format(minutes, seconds)

    cpdef handle_input(self,Input new_input, bint visible, bint parent_read_only = False):
        self.backwards.handle_input(new_input, True, parent_read_only=False)
        self.play.handle_input(new_input, True, parent_read_only=False)
        self.forwards.handle_input(new_input, True, parent_read_only=False)

        global should_redraw_overlay
        if self.seeking and new_input.dm:
            val = clampmap(new_input.m.x-self.bar.org.x, 0, self.bar.size.x,
                           self.min_ts, self.max_ts)
            self.playback_time_setter(val)
            should_redraw_overlay = True
        elif self.trimming_right and new_input.dm:
            val = clampmap(new_input.m.x-self.bar.org.x, 0, self.bar.size.x,
                           self.min_ts, self.max_ts)
            self.trim_right.value = val
            should_redraw_overlay = True
        elif self.trimming_left and new_input.dm:
            val = clampmap(new_input.m.x-self.bar.org.x, 0, self.bar.size.x,
                           self.min_ts, self.max_ts)
            self.trim_left.value = val
            should_redraw_overlay = True

        if self.seek_handle.mouse_over(new_input.m) or self.seeking:
            should_redraw_overlay = should_redraw_overlay or self.hovering != 1
            self.hovering = 1
        elif self.trim_right_handle.mouse_over(new_input.m) or self.trimming_right:
            should_redraw_overlay = should_redraw_overlay or self.hovering != 2
            self.hovering = 2
        elif self.trim_left_handle.mouse_over(new_input.m) or self.trimming_left:
            should_redraw_overlay = should_redraw_overlay or self.hovering != 3
            self.hovering = 3
        elif self.bar.mouse_over_margin(new_input.m, Vec2(0, 10 * ui_scale)):
            self.point_click_seek_loc = Vec2(new_input.m.x, self.bar.center[1])
            should_redraw_overlay = should_redraw_overlay or self.hovering != 4 or new_input.dm.x != 0
            self.hovering = 4
        else:
            should_redraw_overlay = should_redraw_overlay or self.hovering != 0
            self.hovering = 0


        for b in new_input.buttons[:]: # list copy for remove to work
            if b[1] == 1:
                if self.hovering == 4:
                    val = clampmap(new_input.m.x-self.bar.org.x, 0,
                                   self.bar.size.x, self.min_ts, self.max_ts)
                    self.seeking = True
                    self.seeking_cb(True)
                    self.playback_time_setter(val)
                    self.hovering = 1
                    should_redraw_overlay = True

                if self.hovering == 1:
                    new_input.buttons.remove(b)
                    should_redraw_overlay = True
                    if not self.seeking:  # do not call seeking_cb twice
                        self.seeking = True
                        self.seeking_cb(True)
                elif self.hovering == 2:
                    new_input.buttons.remove(b)
                    self.trimming_right = True
                    should_redraw_overlay = True
                elif self.hovering == 3:
                    new_input.buttons.remove(b)
                    self.trimming_left = True
                    should_redraw_overlay = True

            if self.seeking and b[1] == 0:
                self.seeking = False
                should_redraw_overlay = True
                self.seeking_cb(False)
            elif self.trimming_right and b[1] == 0:
                self.trimming_right = False
                should_redraw_overlay = True
            elif self.trimming_left and b[1] == 0:
                self.trimming_left = False
                should_redraw_overlay = True



# cython: language_level=3

cimport pyglui.cygl.glew as gl


cdef class Shader:
    cdef dict uniforms
    cdef bytes _vertex_code, _fragment_code, _geometry_code
    cdef public gl.GLuint handle
    cdef bint linked

    cdef _build_shader(self, const char * strings, shader_type)

    cdef _link(self)

    cpdef bind(self)

    cpdef unbind(self)

    cpdef uniformi(self, str name, vals)

    cpdef uniformf(self, str name, vals)

    cpdef uniform1f(self, str name,float val)

    cpdef uniform1i(self, str name,int val)

    cpdef uniform_matrixf(self, str name, float[:] mat)

    cdef _get_shader_info(self,gl.GLuint shader)

    cdef _get_program_info(self)



# cython: language_level=3
cimport pyglui.pyfontstash.cfontstash as fs


cdef int FONS_ALIGN_LEFT
cdef int FONS_ALIGN_CENTER
cdef int FONS_ALIGN_RIGHT
cdef int FONS_ALIGN_TOP
cdef int FONS_ALIGN_MIDDLE
cdef int FONS_ALIGN_BOTTOM
cdef int FONS_ALIGN_BASELINE
cdef int FONS_ZERO_TOPLEFT
cdef int FONS_ZERO_BOTTOMLEFT


cdef class Context:
    cdef fs.FONScontext * ctx
    cdef dict fonts

    cpdef draw_text(self,float x,float y ,object text)
    cpdef set_color_float(self,tuple color)

    #custom conviniece methods
    cpdef draw_limited_text(self, float x, float y, object text, float width)
    cpdef get_first_char_idx(self, object text, float width)
    cpdef draw_multi_line_text(self, float x, float y, object text, float line_height =*)
    cpdef compute_breaking_text(self, float x, float y, object text, float width,float height,float line_height =*)
    cpdef draw_breaking_text(self, float x, float y, object text, float width,float height,float line_height =*)
    cpdef vertical_metrics(self)
    cpdef char_cumulative_width(self, float x, float y, object text)



########## Global Design Parameters ##########
DEF text_size = 18.
DEF line_height = 20.
DEF x_spacer = 5
DEF rect_color_default = (.0,.0,.0,.8)


########## UI Elements Design Parameters ##########
# ui element design params should be declared here
# name variables starting with the element name
# example: `slider_` or `text_input_` followed by the attributes

# UI_element parameters - base class
DEF outline_padding = 10
DEF circle_button_size = 20
DEF circle_button_size_selected = 25
DEF circle_button_shadow = 10

DEF color_selected = (0.5, 0.8, 0.75,.9)
DEF color_on = (0.5, 0.8, 0.75,.9)
DEF color_on_read_only = (0.5, 0.8, 0.75,.75)
DEF color_default = (.5,.5,.5,.9)
DEF color_default_read_only = (.5,.5,.5,.75)
DEF color_shadow = (.0,.0,.0,.8)
DEF color_shadow_read_only = (.0,.0,.0,.0)
DEF shadow_sharpness = 0.3
DEF color_text_default = (1.,1.,1.,1.)
DEF color_text_info = (1.,1.,1.,.6)
DEF color_text_read_only = (.5,.5,.5,.5)
DEF color_line_default = (1.,1.,1.,.4)
DEF size_text_info = 18

# Slider - design parameters
DEF slider_outline_size_y = 65.
DEF slider_label_org_y = 20.
DEF slider_handle_org_y = 40.
DEF slider_button_size = circle_button_size
DEF slider_button_size_read_only = 15.
DEF slider_button_size_selected = circle_button_size_selected
DEF slider_button_shadow = circle_button_shadow
DEF slider_step_mark_size = 8.
DEF slider_color_step = (.8,.8,.8,.6)
DEF slider_line_color_default = color_line_default
DEF slider_line_color_highlight = color_on
DEF slider_line_color_default_read_only = color_line_default
DEF slider_line_color_highlight_read_only = color_on_read_only

# Switch - design parameters
DEF switch_outline_size_y = 40
DEF switch_button_size = circle_button_size
DEF switch_button_size_selected = circle_button_size_selected
DEF switch_button_size_on = switch_button_size+5
DEF switch_button_shadow = circle_button_shadow+3

# Selector - design parameters
DEF selector_outline_size_y = 40.
DEF selector_triangle_color_default = color_line_default
DEF selector_triangle_color_read_only = color_text_read_only

# Info_Text - design parameters
#DEF info_text_outline_size = 40

# TextInput - design parameters
DEF text_input_outline_size_y = 40.
DEF text_input_highlight_color = (.5,.5,.9,.5)
DEF text_input_line_highlight_color = slider_line_color_highlight

# Button - design parameters
DEF button_corner_radius = 5.
DEF button_outline_size_y = 44.
DEF button_outline_padding = outline_padding
DEF button_text_padding = 2.
DEF button_default_color = color_default
DEF button_read_only_color = (0.25, 0.25, 0.25, 0.75)
DEF button_active_color = color_selected
DEF button_default_text_color = color_text_default
DEF button_read_only_text_color = (0.5, 0.5, 0.5, 0.75)
DEF button_active_text_color = color_shadow

# Thumb - design parameters
DEF thumb_outline_size = 80.
DEF thumb_outline_pad = 0.
DEF thumb_button_size_offset_on = 30.
DEF thumb_button_size_offset_selected = 25.
DEF thumb_button_size_offset_off = thumb_button_size_offset_on
DEF thumb_color_on = color_on
DEF thumb_color_off = (1,1,1,0.6)
DEF thumb_color_shadow = (.0,.0,.0,.0)
DEF thumb_button_sharpness = 0.8
DEF thumb_button_shadow_sharpness = 0.5
DEF thumb_font_padding = 30.

DEF icon_outline_size = 50.
DEF icon_font_padding = 20.
DEF icon_progress_color = color_on

DEF tooltip_text_size = 18.

# Menu - design parameters
DEF menu_pad = 10.
DEF menu_move_corner_height = text_size
DEF menu_move_corner_width = menu_move_corner_height
DEF menu_topbar_pad = 2 * menu_pad + menu_move_corner_height
DEF menu_topbar_text_x_org =  menu_move_corner_width + x_spacer
DEF resize_corner_size = 20.
DEF menu_topbar_min_width = 300.
DEF menu_bottom_pad = 20.

DEF menu_sidebar_pad = menu_topbar_pad
DEF menu_sidebar_min_height = 20.
DEF menu_line = (1.,1.,1.,.8)

DEF seekbar_trim_color = color_on
DEF seekbar_trim_color_hover = (1., 0.84, 0.4, 0.8)
DEF seekbar_seek_color = color_on
DEF seekbar_seek_color_hover = (1., 0.84, 0.4, 0.8)
DEF seekbar_number_size = 15.

DEF timelines_draggable_size = resize_corner_size
DEF timeline_label_size = text_size
